(* This program is free software; you can redistribute it and/or      *)
(* modify it under the terms of the GNU Lesser General Public License *)
(* as published by the Free Software Foundation; either version 2.1   *)
(* of the License, or (at your option) any later version.             *)
(*                                                                    *)
(* This program is distributed in the hope that it will be useful,    *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of     *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      *)
(* GNU General Public License for more details.                       *)
(*                                                                    *)
(* You should have received a copy of the GNU Lesser General Public   *)
(* License along with this program; if not, write to the Free         *)
(* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA *)
(* 02110-1301 USA                                                     *)



Require Import Termes.
Require Import Conv.
Require Import Types.
Require Import Class.

  (* schemes of reducibility: functions on sets of terms *)  

  Fixpoint Can (K : skel) : Type :=
    match K with
    | PROP => term -> Prop
    | PROD s1 s2 => Can s1 -> Can s2
    end.


  (* equality on schemes *)

  Definition eq_cand (X Y : term -> Prop) : Prop :=
    forall t : term, X t <-> Y t.


  Hint Unfold eq_cand: coc.

  Fixpoint eq_can (s : skel) : Can s -> Can s -> Prop :=
    match s as s0 return (Can s0 -> Can s0 -> Prop) with
    | PROP => eq_cand
    | PROD s1 s2 =>
        fun C1 C2 : Can (PROD s1 s2) =>
        forall X1 X2 : Can s1,
        eq_can s1 X1 X2 ->
        eq_can s1 X1 X1 -> eq_can s1 X2 X2 -> eq_can s2 (C1 X1) (C2 X2)
    end.

  Hint Unfold iff: coc.



  Lemma eq_can_sym :
   forall (s : skel) (X Y : Can s), eq_can s X Y -> eq_can s Y X.
simple induction s; simpl in |- *; intros; auto with coc core arith datatypes.
unfold eq_cand in |- *; intros.
elim H with t; auto with coc core arith datatypes.
Qed.

  Lemma eq_can_trans :
   forall (s : skel) (a b c : Can s),
   eq_can s a b -> eq_can s b b -> eq_can s b c -> eq_can s a c.
simple induction s; simpl in |- *; intros.
unfold eq_cand in |- *; intros.
elim H with t; elim H1 with t; auto with coc core arith datatypes.

apply H0 with (b X1); auto with coc core arith datatypes.
Qed.


  Lemma eq_cand_incl :
   forall (t : term) (X Y : Can PROP), eq_can PROP X Y -> X t -> Y t.
intros.
elim H with t; auto with coc core arith datatypes.
Qed.



  (* Higher order candidates of reducibility *)

  Inductive neutral : term -> Prop := 
    | neutral_ref v : neutral (Ref v) 
    | neutral_app u v : neutral (App u v)
    | neutral_prod T U : neutral (Prod T U)
    | neutral_sigma T U : neutral (Sigma T U)
    | neutral_proj1 M : neutral (Proj1 M)
    | neutral_proj2 M : neutral (Proj2 M).

  Hint Immediate neutral_ref neutral_app neutral_prod neutral_sigma neutral_proj1 neutral_proj2 : coc.


  Record is_cand (X : term -> Prop) : Prop := 
    {incl_sn : forall t : term, X t -> sn t;
     clos_red : forall t : term, X t -> forall u : term, red1 t u -> X u;
     clos_exp :
      forall t : term, neutral t -> (forall u : term, red1 t u -> X u) -> X t}. 



  Lemma var_in_cand :
   forall (n : nat) (X : term -> Prop), is_cand X -> X (Ref n).
intros.
apply (clos_exp X); auto with coc core arith datatypes.

intros.
inversion H0.
Qed.


  Lemma clos_red_star :
   forall R : term -> Prop,
   is_cand R -> forall a b : term, R a -> red a b -> R b.
simple induction 3; intros; auto with coc core arith datatypes.
apply (clos_red R) with P; auto with coc core arith datatypes.
Qed.


  Lemma cand_sat :
   forall X : term -> Prop,
   is_cand X ->
   forall T : term,
   sn T ->
   forall u : term,
   sn u -> forall m : term, X (subst u m) -> X (App (Abs T m) u).
unfold sn in |- *.
simple induction 2.
simple induction 3.
intros.
generalize H6.
elimtype (sn m); intros.
apply (clos_exp X); intros; auto with coc core arith datatypes.

inversion_clear H10; auto with coc core arith datatypes.
inversion_clear H11.
apply H2; auto with coc core arith datatypes.
apply Acc_intro; auto with coc core arith datatypes.

apply H8; auto with coc core arith datatypes.
apply (clos_red X) with (subst x0 x1); auto with coc core arith datatypes.
unfold subst in |- *; auto with coc core arith datatypes.

apply H5; auto with coc core arith datatypes.
apply clos_red_star with (subst x0 x1); auto with coc core arith datatypes.
unfold subst in |- *; auto with coc core arith datatypes.

apply sn_subst with x0.
apply (incl_sn X); auto with coc core arith datatypes.
Qed.



  Fixpoint is_can (s : skel) : Can s -> Prop :=
    match s as s0 return (Can s0 -> Prop) with
    | PROP => fun X : term -> Prop => is_cand X
    | PROD s1 s2 =>
        fun C : Can s1 -> Can s2 =>
        forall X : Can s1, is_can s1 X -> eq_can s1 X X -> is_can s2 (C X)
    end.


  Lemma is_can_prop : forall X : term -> Prop, is_can PROP X -> is_cand X.
auto with coc core arith datatypes.
Qed.

  Hint Resolve is_can_prop: coc.




  (* Default Candidates *)

  Fixpoint default_can (s : skel) : Can s :=
    match s as ss return (Can ss) with
    | PROP => sn
    | PROD s1 s2 => fun _ : Can s1 => default_can s2
    end.


  Lemma cand_sn : is_cand sn.
apply Build_is_cand; intros; auto with coc core arith datatypes.

apply sn_red_sn with t; auto with coc core arith datatypes.

red in |- *; apply Acc_intro; auto with coc core arith datatypes.
Qed.

  Hint Resolve cand_sn: coc.


  Lemma def_can_cr : forall s : skel, is_can s (default_can s).
simple induction s; simpl in |- *; intros; auto with coc core arith datatypes.
Qed.


  Lemma def_inv : forall s : skel, eq_can s (default_can s) (default_can s).
simple induction s; simpl in |- *; intros; auto with coc core arith datatypes.
Qed.


  Hint Resolve def_inv def_can_cr: coc.



  Definition Pi (s : skel) (X : term -> Prop) (F : Can (PROD s PROP))
    (t : term) : Prop :=
    forall u : term,
    X u -> forall C : Can s, is_can s C -> eq_can s C C -> F C (App t u).


  Lemma eq_can_Pi :
   forall (s : skel) (X Y : term -> Prop) (F1 F2 : Can (PROD s PROP)),
   eq_can PROP X Y ->
   eq_can (PROD s PROP) F1 F2 -> eq_can PROP (Pi s X F1) (Pi s Y F2).
simpl in |- *; intros; unfold iff, Pi in |- *.
split; intros.
elim H0 with C C (App t u); elim H with u; auto with coc core arith datatypes.

elim H0 with C C (App t u); elim H with u; auto with coc core arith datatypes.
Qed.



  Lemma is_can_Pi :
   forall (s : skel) (X : term -> Prop),
   is_cand X ->
   forall F : Can (PROD s PROP), is_can (PROD s PROP) F -> is_cand (Pi s X F).
simpl in |- *; unfold Pi in |- *; intros.
apply Build_is_cand; intros.
apply subterm_sn with (App t (Ref 0)); auto with coc core arith datatypes.
apply (incl_sn (F (default_can s))); auto with coc core arith datatypes.
apply H1; auto with coc core arith datatypes.
apply (var_in_cand 0 X); auto with coc core arith datatypes.

apply (clos_red (F C)) with (App t u0); auto with coc core arith datatypes.

apply (clos_exp (F C)); auto with coc core arith datatypes.

generalize H3.
cut (sn u).
simple induction 1; intros.
generalize H1.
inversion_clear H10; intros; auto with coc core arith datatypes.
inversion H10.

apply (clos_exp (F C)); intros; auto with coc core arith datatypes.

apply H8 with N2; auto with coc core arith datatypes.
apply (clos_red X) with x; auto with coc core arith datatypes.

apply (incl_sn X); auto with coc core arith datatypes.
Qed.


  Definition Cartesian (X Y : Can PROP)
    (t : term) : Prop :=
    X (Proj1 t) /\ Y (Proj2 t).

  Lemma cartesian_eq_can X1 Y1 X2 Y2 : eq_can PROP X1 X2 -> eq_can PROP Y1 Y2 -> eq_can PROP (Cartesian X1 Y1) (Cartesian X2 Y2).
intros eqX eqY.
split ; intros [cX cY] ;
  ( split ; [ now apply eqX | now apply eqY ] ).
Qed.

  Lemma cartesian_is_cand X Y : is_can PROP X -> is_can PROP Y -> is_can PROP (Cartesian X Y).
intros cX cY.
destruct cX.
destruct cY.
constructor.
  intros t [c1tX c2tY].
  apply subterm_sn with (Proj1 t) ; superauto.
  
  intros t [c1tX c2tY] u rtu.
  constructor ; superauto.
    apply clos_red0 with (Proj1 t) ; superauto.
    
    apply clos_red1 with (Proj2 t) ; superauto.
  
  intros t nt all_red1_cXY.
  constructor ;
    [ apply clos_exp0
    | apply clos_exp1 ] ; superauto ;
    ( intros u r ; inversion r ; subst ; superauto ;
      [ inversion nt
      | apply all_red1_cXY ] ; superauto ).
Qed.

  Lemma cartesian_closure X Y : is_can PROP X -> eq_can PROP X X -> is_can PROP Y -> eq_can PROP Y Y -> is_cand (Cartesian X Y) /\ eq_can PROP (Cartesian X Y) (Cartesian X Y).

intros cX eqX cY eqY.
split.
  apply cartesian_is_cand ; superauto.
  
  apply cartesian_eq_can ; superauto.
Qed.

  Lemma cartesian_intro A B M N X Y : is_cand X -> is_cand Y -> sn A -> sn B -> X M -> Y N -> Cartesian X Y (Pair A B M N).
intros cX cY snA snB xM yN.
assert (H : sn (Pair A B M N)).
  apply sn_pair ; superauto.
  apply (incl_sn X) ; superauto.
  apply (incl_sn Y) ; superauto.
remember (Pair A B M N) as x.
clear snA snB.
revert A B M N Heqx xM yN.
induction H.
intros A B M N eqx xM yN.

subst.
(* write it down nicely... hehehe *)
assert (forall A' B' M' N', red1 (Pair A B M N) (Pair A' B' M' N') -> X M' -> Y N' -> Cartesian X Y (Pair A' B' M' N')).
  intros A' B' M' N' r'. apply H0 with A' B' ; superauto.
clear H H0.

(* now we can get to work :) *)
constructor ; 
  [ apply (clos_exp X)
  | apply (clos_exp Y) ] ; superauto ;
    intros u r ;
    inversion r ; subst ; superauto ; (
      inversion_clear H0 ; apply H1 ; superauto ; (* I'm a genius aren't I? *)
        [ apply (clos_red X) with M 
        | apply (clos_red Y) with N ] ; superauto ).
Qed.

  Lemma cartesian_elim M X Y : is_cand X -> is_cand Y -> Cartesian X Y M -> forall P : Prop, (X (Proj1 M) -> Y (Proj2 M) -> P) -> P.
intros.
apply H2 ; apply H1.
Qed.

  Lemma Abs_sound :
   forall (A : term -> Prop) (s : skel) (F : Can s -> term -> Prop)
     (T m : term),
   is_can PROP A ->
   is_can (PROD s PROP) F ->
   (forall n : term,
    A n -> forall C : Can s, is_can s C -> eq_can s C C -> F C (subst n m)) ->
   sn T -> Pi s A F (Abs T m).
unfold Pi in |- *; simpl in |- *; intros.
cut (is_cand (F C)); intros; auto with coc core arith datatypes.
apply (clos_exp (F C)); intros; auto with coc core arith datatypes.

apply clos_red with (App (Abs T m) u); auto with coc core arith datatypes.
apply (cand_sat (F C)); auto with coc core arith datatypes.
apply (incl_sn A); auto with coc core arith datatypes.
Qed.

  Lemma simpl_eq_cand X : eq_can PROP X X.
split ; superauto.
Qed.

Hint Resolve simpl_eq_cand : coc.

